# -*- coding: utf-8 -*-
#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import logging
from typing import Any, Dict, Iterable, Type

from .base import KelvinObject, KelvinResource
from .session import Session

logger = logging.getLogger(__name__)


class Role(KelvinObject):
    _class_display_name = "Role"
    _kelvin_attrs = ["name", "display_name"]

    def __init__(
        self,
        name: str,
        *,
        display_name: str = None,
        url: str = None,
        session: Session = None,
    ):
        super().__init__(
            name=name, ucsschool_roles=None, udm_properties=None, dn=None, url=url, session=session
        )
        self.display_name = display_name
        del self.dn
        del self.ucsschool_roles
        del self.udm_properties
        self._resource_class = RoleResource

    async def save(self) -> "RoleResource":
        raise NotImplementedError(
            "Creating and changing of school roles has not yet been "
            "implemented in the Kelvin REST API."
        )

    async def delete(self) -> None:
        raise NotImplementedError(
            "Deleting school roles has not yet been implemented in the " "Kelvin REST API."
        )

    def as_dict(self) -> Dict[str, Any]:
        attrs = self._kelvin_attrs + ["url"]
        return dict((attr, getattr(self, attr)) for attr in attrs)


class RoleResource(KelvinResource):
    class Meta:
        kelvin_object: Type[KelvinObject] = Role
        required_get_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ()

    def __init__(self, session: Session):
        super().__init__(session=session)
        self.collection_url = self.session.urls["role"]
        self.object_url = f"{self.session.urls['role']}{{name}}"



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import sys
import warnings

import pytest
from async_property import async_property

from ucsschool.kelvin.client.session import BadSettingsWarning, Session

_PYTHON_VERSION = sys.version_info[:2]
kelvin_session_kwargs_mock = {
    "username": "username",
    "password": "password",
    "host": "localhost",
    "verify": False,
}


class SessionMock:
    @async_property
    async def token(self) -> str:
        return "Token"


def call_kwargs(call_args):
    if _PYTHON_VERSION < (3, 8):
        return call_args[1]
    else:
        return call_args.kwargs


@pytest.mark.asyncio
async def test_session_closes_on_context_exit():
    async with Session(**kelvin_session_kwargs_mock) as session:
        assert session._client
    with pytest.raises(RuntimeError):
        print(session.client)


@pytest.mark.asyncio
@pytest.mark.parametrize("arg_client_tasks", range(-10, 11))
async def test_session_warn_max_client_tasks(arg_client_tasks):
    with warnings.catch_warnings(record=True) as w:
        async with Session(max_client_tasks=arg_client_tasks, **kelvin_session_kwargs_mock) as session:
            assert session.max_client_tasks >= 4
        if arg_client_tasks < 4:
            assert len(w) == 1
            assert issubclass(w[-1].category, BadSettingsWarning)
            assert "max_client_tasks" in str(w[-1].message)


@pytest.mark.asyncio
async def test_session_default_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    async with Session(**kelvin_session_kwargs_mock) as session:
        try:
            await session.get("http://example.com")
        except NotImplementedError:
            pass
        assert call_kwargs(session.client.get.call_args)["timeout"] == 10.0


@pytest.mark.asyncio
async def test_session_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = dict(kelvin_session_kwargs_mock, timeout=5)
    async with Session(**kelvin_session_kwargs) as session:
        try:
            await session.get("http://example.com")
        except NotImplementedError:
            pass
        assert call_kwargs(session.client.get.call_args)["timeout"] == 5


@pytest.mark.asyncio
async def test_session_explicit_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    async with Session(**kelvin_session_kwargs_mock) as session:
        try:
            await session.get("http://example.com", timeout=7)
        except NotImplementedError:
            pass
        assert call_kwargs(session.client.get.call_args)["timeout"] == 7


@pytest.mark.asyncio
async def test_session_override_timeout(mocker):
    mocker.patch("httpx.AsyncClient.get", side_effect=NotImplementedError)
    mocker.patch("ucsschool.kelvin.client.session.Session.token", SessionMock.token)
    kelvin_session_kwargs = dict(kelvin_session_kwargs_mock, timeout=4)
    async with Session(**kelvin_session_kwargs) as session:
        try:
            await session.get("http://example.com", timeout=6)
        except NotImplementedError:
            pass
        assert call_kwargs(session.client.get.call_args)["timeout"] == 6



# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

from dataclasses import asdict

import ldap3
import pytest
from faker import Faker

from ucsschool.kelvin.client import School, SchoolResource, Session

fake = Faker()


API_VERSION = "v1"
URL_BASE = "https://{host}/ucsschool/kelvin"
URL_TOKEN = f"{URL_BASE}/token"
URL_SCHOOL_RESOURCE = f"{URL_BASE}/{API_VERSION}/schools/"
URL_SCHOOL_COLLECTION = URL_SCHOOL_RESOURCE
URL_SCHOOL_OBJECT = f"{URL_SCHOOL_RESOURCE}{{name}}"


@pytest.mark.asyncio
async def test_search_no_name_arg(compare_kelvin_obj_with_test_data, new_school, kelvin_session_kwargs):
    school1, school2 = new_school(2)

    async with Session(**kelvin_session_kwargs) as session:
        objs = [obj async for obj in SchoolResource(session=session).search()]

    assert objs, "No Schools found."
    assert len(objs) >= 2
    assert school1.dn in [school.dn for school in objs]
    assert school2.dn in [school.dn for school in objs]
    for obj in objs:
        if obj.dn == school1.dn:
            compare_kelvin_obj_with_test_data(obj, **asdict(school1))
        if obj.dn == school2.dn:
            compare_kelvin_obj_with_test_data(obj, **asdict(school2))


@pytest.mark.asyncio
async def test_search_partial_name_arg(
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    new_school,
):
    school = new_school(1)[0]
    name_len = len(school.name)
    name_begin = school.name[: int(name_len / 2)]
    name_end = school.name[len(name_begin) :]

    async with Session(**kelvin_session_kwargs) as session:
        objs1 = [obj async for obj in SchoolResource(session=session).search(name=f"{name_begin}*")]
        objs2 = [obj async for obj in SchoolResource(session=session).search(name=f"*{name_end}")]
    assert objs1, f"No School for name='{name_begin}*' found."
    assert len(objs1) >= 1
    assert school.dn in [o.dn for o in objs1]
    assert objs2, f"No School for name='*{name_end}' found."
    assert len(objs2) >= 1
    assert school.dn in [o.dn for o in objs2]


@pytest.mark.asyncio
async def test_get_from_url(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, new_school):
    school = new_school(1)[0]
    url = URL_SCHOOL_OBJECT.format(host=kelvin_session_kwargs["host"], name=school.name)
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolResource(session=session).get_from_url(url)
    compare_kelvin_obj_with_test_data(obj, **asdict(school))


@pytest.mark.asyncio
async def test_get(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, new_school):
    school = new_school(1)[0]
    async with Session(**kelvin_session_kwargs) as session:
        obj = await SchoolResource(session=session).get(name=school.name)
    compare_kelvin_obj_with_test_data(obj, **asdict(school))


@pytest.mark.asyncio
async def test_create_minimal(
    base_dn,
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    ldap_access,
    new_school_test_obj,
    schedule_delete_ou_using_ssh,
):
    school_data = new_school_test_obj()
    schedule_delete_ou_using_ssh(school_data.name)
    async with Session(**kelvin_session_kwargs) as session:
        school_kwargs = {"name": school_data.name}
        print(f"Creating school with kwargs: {school_kwargs!r}")
        school_obj = School(session=session, **school_kwargs)
        await school_obj.save()
        print("Created new School: {!r}".format(school_obj.as_dict()))

    ldap_filter = f"(&(ou={school_obj.name})(objectClass=ucsschoolOrganizationalUnit))"
    ldap_objs = await ldap_access.search(filter_s=ldap_filter)
    assert len(ldap_objs) == 1
    ldap_obj = ldap_objs[0]
    assert ldap_obj.entry_dn == school_obj.dn
    assert ldap_obj["ou"].value == school_obj.name
    assert ldap_obj["ucsschoolRole"].value == f"school:school:{school_obj.name}"
    assert ldap_obj["displayName"].value == school_obj.name
    dc_base_dn = f"cn=dc,cn=server,cn=computers,ou={school_obj.name},{base_dn}"
    class_share_file_server_dns = [f"cn=dc{school_obj.name},{dc_base_dn}"]
    home_share_file_server_dns = [f"cn=dc{school_obj.name},{dc_base_dn}"]
    single_server_objs = await ldap_access.search(
        filter_s=f"(ucsschoolRole=single_master:school:{school_obj.name})", attributes=["cn"]
    )
    if single_server_objs:
        single_server_cn = single_server_objs[0]["cn"].value
        single_server_dn = f"cn={single_server_cn},cn=dc,cn=computers,{base_dn}"
        class_share_file_server_dns.append(single_server_dn)
        home_share_file_server_dns.append(single_server_dn)
    assert ldap_obj["ucsschoolClassShareFileServer"].value in class_share_file_server_dns
    assert ldap_obj["ucsschoolHomeShareFileServer"].value in home_share_file_server_dns
    # no need to check the groups, we're testing the client, not the Kelvin API or ucsschool.lib


@pytest.mark.asyncio
async def test_create_all_attrs(
    base_dn,
    compare_kelvin_obj_with_test_data,
    kelvin_session_kwargs,
    ldap_access,
    new_school_test_obj,
    schedule_delete_ou_using_ssh,
):
    school_data = new_school_test_obj()
    schedule_delete_ou_using_ssh(school_data.name)
    async with Session(**kelvin_session_kwargs) as session:
        school_kwargs = asdict(school_data)
        school_kwargs["udm_properties"] = {"description": fake.first_name()}
        print(f"Creating school with kwargs: {school_kwargs!r}")
        school_obj = School(session=session, **school_kwargs)
        await school_obj.save()
        print("Created new School: {!r}".format(school_obj.as_dict()))

    ldap_filter = f"(&(ou={school_obj.name})(objectClass=ucsschoolOrganizationalUnit))"
    ldap_objs = await ldap_access.search(filter_s=ldap_filter)
    assert len(ldap_objs) == 1
    ldap_obj = ldap_objs[0]
    assert ldap_obj.entry_dn == school_obj.dn
    assert ldap_obj["ou"].value == school_obj.name
    assert ldap_obj["ucsschoolRole"].value == f"school:school:{school_obj.name}"
    assert ldap_obj["displayName"].value == school_obj.display_name
    dc_base_dn = f"cn=dc,cn=server,cn=computers,ou={school_obj.name},{base_dn}"
    class_share_file_server_dn_dn = f"cn={school_obj.class_share_file_server},{dc_base_dn}"
    home_share_file_server_dn = f"cn={school_obj.home_share_file_server},{dc_base_dn}"
    assert ldap_obj["ucsschoolClassShareFileServer"].value == class_share_file_server_dn_dn
    assert ldap_obj["ucsschoolHomeShareFileServer"].value == home_share_file_server_dn
    assert ldap_obj["description"].value == school_kwargs["udm_properties"]["description"]
    # no need to check the groups, we're testing the client, not the Kelvin API or ucsschool.lib


@pytest.mark.asyncio
async def test_reload(compare_kelvin_obj_with_test_data, kelvin_session_kwargs, ldap_access, new_school):
    school = new_school(1)[0]
    display_name_old = school.display_name
    display_name_new = fake.text(max_nb_chars=50)

    async with Session(**kelvin_session_kwargs) as session:
        obj: School = await SchoolResource(session=session).get(name=school.name)
        assert obj.display_name == display_name_old
        await obj.reload()
        assert obj.display_name == display_name_old
        await ldap_access.modify(obj.dn, {"displayName": [(ldap3.MODIFY_REPLACE, [display_name_new])]})
        await obj.reload()
        assert obj.display_name == display_name_new
        await ldap_access.modify(obj.dn, {"displayName": [(ldap3.MODIFY_REPLACE, [display_name_old])]})



# Copyright 2022 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import logging

import pytest
from faker import Faker

from ucsschool.kelvin.client import SchoolResource, Session

fake = Faker()
logger = logging.getLogger(__name__)


async def search_school(new_school, kelvin_session_kwargs):
    new_school(1)

    async with Session(**kelvin_session_kwargs) as session:
        SchoolResource(session=session).search()


@pytest.mark.parametrize(
    "log_level",
    [
        pytest.param(logging.CRITICAL, id="log_CRITICAL"),
        pytest.param(logging.ERROR, id="log_ERROR"),
        pytest.param(logging.WARNING, id="log_WARNING"),
        pytest.param(logging.INFO, id="log_INFO"),
        pytest.param(logging.DEBUG, id="log_DEBUG"),
        pytest.param(logging.NOTSET, id="log_NOTSET"),
    ],
)
@pytest.mark.asyncio
async def test_log_mask_credentials(caplog, new_school, kelvin_session_kwargs, log_level):
    filters = {
        "Authorization": "'Authorization': '**********'",
        "username": f"'username': '{kelvin_session_kwargs['username']}'",
        "password": "'password': '**********'",
    }

    with caplog.at_level(log_level):
        await search_school(new_school, kelvin_session_kwargs)
        for line in caplog.text.split("\n"):
            for key, value in filters.items():
                if key in line:
                    assert value in line
                # exclude the username from masking
                if key in kelvin_session_kwargs and key != "username":
                    assert kelvin_session_kwargs[key] not in line



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup

with open("README.rst") as fp:
    readme = fp.read()
with open("HISTORY.rst") as fp:
    history = fp.read()
with open("requirements.txt") as fp:
    requirements = fp.read().splitlines()
with open("requirements_dev.txt") as fp:
    requirements_dev = fp.read().splitlines()
with open("requirements_test.txt") as fp:
    requirements_test = fp.read().splitlines()
with open("VERSION.txt") as fp:
    version = fp.read().strip()

setup(
    author="Daniel Troeder",
    author_email="troeder@univention.de",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ],
    description="Python library to interact with the UCS@school Kelvin REST API.",
    license="GNU Affero General Public License v3",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="Univention UCS UCS@school REST",
    name="kelvin-rest-api-client",
    packages=["ucsschool.kelvin.client"],
    install_requires=requirements,
    # setup_requires=["pytest-runner"],
    test_suite="tests",
    tests_require=requirements_dev + requirements_test,
    extras_require={"development": set(requirements + requirements_dev + requirements_test)},
    python_requires=">=3.7",
    url="https://github.com/univention/kelvin-rest-api-client",
    version=version,
)
